@@ -53,11 +53,23 @@ static INTRINSICS0: phf::Map<&'static str, NamedFn0> = phf_map! {
 
     "nan" => NamedFn0 { name: "nan", func: math::nan },
     "infinite" => NamedFn0 { name: "infinite", func: math::infinite },
+    "isnan" => as_math_fn!(is_nan),
+    "isnormal" => as_math_fn!(is_normal),
+    "isfinite" => as_math_fn!(is_finite),
     "floor" => as_math_fn!(floor),
     "sqrt" => as_math_fn!(sqrt),
-    "isnan" => as_math_fn!(is_nan),
-    "isfinite" => as_math_fn!(is_finite),
-    "isnormal" => as_math_fn!(is_normal),
+    "sin" => as_math_fn!(sin),
+    "cos" => as_math_fn!(cos),
+    "tan" => as_math_fn!(tan),
+    "asin" => as_math_fn!(asin),
+    "acos" => as_math_fn!(acos),
+    "atan" => as_math_fn!(atan),
+    "sinh" => as_math_fn!(sinh),
+    "cosh" => as_math_fn!(cosh),
+    "tanh" => as_math_fn!(tanh),
+    "asinh" => as_math_fn!(asinh),
+    "acosh" => as_math_fn!(acosh),
+    "atanh" => as_math_fn!(atanh),
 };
 static INTRINSICS1: phf::Map<&'static str, NamedFn1> = phf_map! {
     "error" => NamedFn1 { name: "error", func: error1 },
