@@ -9,7 +9,7 @@ use crate::{
     intrinsic,
     util::make_owned,
     vm::{
-        bytecode::{ClosureAddress, Label, NamedFunction},
+        bytecode::{ClosureAddress, NamedFunction},
         error::QueryExecutionError,
         Address, ByteCode, Program, Result, ScopeId, ScopedSlot, Value,
     },
@@ -98,14 +98,18 @@ impl Iterator for PathValueIterator {
 type Frames = PVector<Option<Frame>>;
 type Closure = (ClosureAddress, Frames);
 
+#[derive(Debug, Copy, Clone, PartialEq, Eq)]
+pub struct LabelId(usize);
+
 #[derive(Debug, Clone)]
 struct Frame {
     slots: Rc<RefCell<Vec<Option<Value>>>>,
     closure_slots: Rc<RefCell<Vec<Option<Closure>>>>,
+    label_slots: Rc<RefCell<Vec<Option<LabelId>>>>,
 }
 
 impl Frame {
-    fn new(variable_cnt: usize, closure_cnt: usize) -> Self {
+    fn new(variable_cnt: usize, closure_cnt: usize, label_cnt: usize) -> Self {
         Self {
             slots: Rc::new(RefCell::new(
                 std::iter::repeat(None).take(variable_cnt).collect(),
@@ -113,6 +117,9 @@ impl Frame {
             closure_slots: Rc::new(RefCell::new(
                 std::iter::repeat(None).take(closure_cnt).collect(),
             )),
+            label_slots: Rc::new(RefCell::new(
+                std::iter::repeat(None).take(label_cnt).collect(),
+            )),
         }
     }
 }
@@ -124,6 +131,7 @@ pub struct Machine {
 
 #[derive(Debug)]
 struct Environment {
+    next_label_id: usize,
     forks: Vec<(<State as Undo>::UndoToken, OnFork)>,
 }
 
@@ -148,13 +156,14 @@ enum OnFork {
     CatchError,
     SkipCatch,
     TryAlternative,
-    CatchLabel(Label),
+    CatchLabel(LabelId),
     Iterate,
 }
 
 impl Environment {
     fn new(state: <State as Undo>::UndoToken) -> Self {
         Self {
+            next_label_id: 0,
             forks: vec![(state, OnFork::Nop)],
         }
     }
@@ -169,6 +178,12 @@ impl Environment {
     fn pop_fork(&mut self) -> Option<(<State as Undo>::UndoToken, OnFork)> {
         self.forks.pop()
     }
+
+    fn gen_label_id(&mut self) -> LabelId {
+        let ret = LabelId(self.next_label_id);
+        self.next_label_id += 1;
+        ret
+    }
 }
 
 impl State {
@@ -305,12 +320,29 @@ impl State {
         })
     }
 
+    fn label_slot(&mut self, scoped_slot: &ScopedSlot) -> RefMut<Option<LabelId>> {
+        let frame = self
+            .frames
+            .get_mut(scoped_slot.0 .0)
+            .and_then(Option::as_mut)
+            .ok_or(ProgramError::UninitializedFrame)
+            .unwrap();
+        let slots = frame.label_slots.borrow_mut();
+        RefMut::map(slots, |v| {
+            v.get_mut(scoped_slot.1)
+                .ok_or(ProgramError::UnknownSlot)
+                .unwrap()
+        })
+    }
+
+    #[allow(clippy::too_many_arguments)]
     fn push_frame(
         &mut self,
         context_frames: Option<Frames>,
         scope_id: ScopeId,
         variable_cnt: usize,
         closure_cnt: usize,
+        label_cnt: usize,
         chain_ret: bool,
         return_address: Address,
     ) {
@@ -323,7 +355,7 @@ impl State {
             self.frames
                 .extend(std::iter::repeat(None).take(scope_id.0 - self.frames.len() + 1))
         }
-        self.frames[scope_id.0] = Some(Frame::new(variable_cnt, closure_cnt));
+        self.frames[scope_id.0] = Some(Frame::new(variable_cnt, closure_cnt, label_cnt));
         self.frame_stack.push((return_address, saved, chain_ret));
     }
 
@@ -471,9 +503,9 @@ fn run_code(program: &Program, state: &mut State, env: &mut Environment) -> Opti
                     if catch_skip == 0 {
                         match err.take() {
                             None => continue 'select_fork,
-                            Some(QueryExecutionError::UserDefinedError(s)) => {
+                            Some(QueryExecutionError::UserDefinedError(v)) => {
                                 state.undo(token);
-                                state.push(Value::string(s));
+                                state.push(v);
                                 break 'select_fork state.save();
                             }
                             Some(e) => {
@@ -729,12 +761,19 @@ fn run_code(program: &Program, state: &mut State, env: &mut Environment) -> Opti
                 },
                 ForkTryEnd => env.push_fork(state, OnFork::SkipCatch, state.pc.get_next()),
                 ForkAlt { fork_pc } => env.push_fork(state, OnFork::TryAlternative, *fork_pc),
-                ForkLabel(label) => {
-                    env.push_fork(state, OnFork::CatchLabel(*label), state.pc.get_next());
+                ForkLabel(label_slot) => {
+                    let label_id = env.gen_label_id();
+                    state.label_slot(label_slot).replace(label_id);
                     // pc doesn't really matter
+                    env.push_fork(state, OnFork::CatchLabel(label_id), state.pc.get_next());
                 }
-                Break(label) => {
-                    err = Some(QueryExecutionError::Breaking(*label));
+                Break(label_slot) => {
+                    let label_id = *state
+                        .label_slot(label_slot)
+                        .as_ref()
+                        .ok_or(ProgramError::UninitializedSlot)
+                        .unwrap();
+                    err = Some(QueryExecutionError::Breaking(label_id));
                     continue 'backtrack;
                 }
                 Backtrack => continue 'backtrack,
@@ -799,6 +838,7 @@ fn run_code(program: &Program, state: &mut State, env: &mut Environment) -> Opti
                     id,
                     variable_cnt,
                     closure_cnt,
+                    label_cnt,
                 } => {
                     let return_address = call_pc
                         .take()
@@ -809,6 +849,7 @@ fn run_code(program: &Program, state: &mut State, env: &mut Environment) -> Opti
                         *id,
                         *variable_cnt,
                         *closure_cnt,
+                        *label_cnt,
                         chain_ret,
                         return_address,
                     );
@@ -828,6 +869,9 @@ fn run_code(program: &Program, state: &mut State, env: &mut Environment) -> Opti
                     log::trace!("Calling function {} with context {:?}", name, context);
                     match func(context) {
                         Ok(value) => state.push(value),
+                        Err(QueryExecutionError::UserDefinedError(Value::Null)) => {
+                            continue 'backtrack
+                        }
                         Err(e) => err = Some(e),
                     }
                 }
@@ -842,6 +886,9 @@ fn run_code(program: &Program, state: &mut State, env: &mut Environment) -> Opti
                     );
                     match func(context, arg1) {
                         Ok(value) => state.push(value),
+                        Err(QueryExecutionError::UserDefinedError(Value::Null)) => {
+                            continue 'backtrack
+                        }
                         Err(e) => err = Some(e),
                     }
                 }
